class Evaluation:

    def generalized_random_forest_confidence_intervals(self): # import as grf_ci
        GRF_CIs = 1
        return GRF_CIs

    def qini_AUUC_scores(self, qini = True):
        qini_AUUC = 1
        return qini_AUUC


  
"""
- The module allows for the download of popular causl inferenfce data
- This allows for quick implementation in a variety of different contexts
- All datasets are cleaned in the loading process

Data at a glance
----------------
1. Kevin Hilstorm's MineThatData email marketing dataset (email marketing)
2. The Primary Biliary Cirrhosis dataset provided by the Mayo Clinic (medical trials)
3. A dataset on microfinance from The Institute for Financial Management Research in Chennai, India (socio-economic)
"""

from .download_utils import download_file, get_download_paths

from .hillstrom import download_hillstrom, load_hillstrom
from .mayo_pbc import download_mayo_pbc, load_mayo_pbc
from .ifmr_microfinance import downlaod_ifmr_microfinance, load_ifmr_microfinance


"""
A dataset on microfinance from The Centre for Micro Finance at the Institute for Financial Management Research (IFMR) (Chennai, India)

Description found at
--------------------
https://www.aeaweb.org/articles?id=10.1257/app.20130533 (see paper)
"""
  
import os
import numpy as np
import pandas as pd
from causeinfer.datasets.download_utils import download_file, get_download_paths

def downlaod_ifmr_microfinance(
    data_path=None,
    url='https://www.aeaweb.org/aej/app/data/0701/2013-0533_data.zip'
):
    """
    Downloads the dataset from the American Economic Assosciation's website.

    Result
    ------
    - A folder with the data in a 'datasets' folder, unless otherwise specified
    """
    data_path, dataset_path = get_download_paths(data_path, 'datasets', 'ifmr_microfinance')
    if not os.path.isdir(data_path):
        os.makedirs(data_path)

    if not os.path.exists(dataset_path):
        download_file(url, dataset_path, zip_file = True)


def __format_data(dataset_path, load_raw_data=False):
    """
    Formats the data upon loading for consistent data preparation.
    Source: https://github.com/thmstang/apa19-microfinance/blob/master/helpers.r
    
    Inputs
    ------
    - The original file is a folder that has various .dta sets

    Returns
    -------
    - A formated version of the data
    """
    # Read in Stata .dta data
    df = pd.read_stata(dataset_path+'/2013-0533_data (TO SUBMIT)/2013-0533_data_endlines1and2.dta')

    if not load_raw_data: # Formats the data if False
    
        # Derive columns for an initial segment based on study baselines
        columns_to_keep = list(df.columns[:15]) # initially select all variables before endline specific variables
        columns_to_keep.extend([col for col in df.columns if col[-len('_1'):] == '_1']) # all variables relevant to Endline1
        columns_to_keep = [col for col in columns_to_keep if col not in ['w', 'w1', 'w2', 'sample1', 'sample2', 'visitday_1', 'visitmonth_1', 'visityear_1']] # exclude survey variables
        columns_to_keep = [col for col in columns_to_keep if col[:len('area_')+1] != 'area_'] # exclude area-level variables
        columns_to_keep = [col for col in columns_to_keep if col[-len('_mo_1'):] != '_mo_1'] # exclude monthly & annually expenses variables (only keep the per capital version)
        columns_to_keep = [col for col in columns_to_keep if col[-len('_annual_1'):] != '_annual_1']

        yes_no_columns = [col for col in columns_to_keep if df[str(col)].isin(['Yes']).any() or df[str(col)].isin(['No']).any()]
        
        df = df[df.columns.intersection(columns_to_keep)]
        df[yes_no_columns] = df[yes_no_columns].eq('Yes').mul(1)
        df['treatment'] = df['treatment'].eq('Treatment').mul(1)

        redundant_cols = [('old_biz', 'any_old_biz'), ('total_biz_1', 'any_biz_1'), ('newbiz_1', 'any_new_biz_1')]

        for tup in redundant_cols:
            mask = list(df[df[tup[1]] == 0].index)
            mask.extend(list(df[df[tup[1]] == np.nan].index))
            df.loc[mask, tup[0]] = 0

        # Removing redundant variables
        redundant_remove = [tup[1] for tup in redundant_cols]
        redundant_remove.extend(['hhsize_1', 'anymfi_1', 'anymfi_amt_1',
                                'anyloan_1', 'anyloan_amt_1', 'hours_week_1', 
                                'hours_headspouse_week_1', 'hours_child1620_week_1', 'total_exp_mo_pc_1'])
        for col in redundant_remove:
            del df[col]

        # Cleaning NaNs - replace, with all remaining NaN rows dropped
        # First remove columns with more than 10% NaN values
        nan_threshold = 0.1
        df = df[df.columns[df.isnull().mean() < nan_threshold]]
        
        # Replce business variables with 0 if total_biz_1 is 0 or NaN
        total_biz_mask = list(df[df['total_biz_1'] == 0].index)
        total_biz_mask.extend(list(df[df['total_biz_1'] == np.nan].index))

        for column in [col for col in df.columns if col[:len('biz')] == 'biz']:
            df.loc[total_biz_mask, column] = 0

        # Fill all non-index variables with their mean
        for column in [col for col in df.columns if 'index' not in col]:
            df[column].fillna(df[column].mean())

        df = df.dropna()

        # Exclude those columns with outliers in the expense related variables
        # Interquartile ranges are used, with 5*iqr times the quartiles being dropped

        exp_col = [col for col in df.columns if 'exp_mo_pc' in col]
        exp_col.extend(['informal_amt_1'])  

        for col in exp_col:
            q75, q25 = np.percentile(df[col], [75 ,25])
            iqr = q75 - q25
            iqr # Here so that VS Code will leave me alone about an unused variable (see @s)
            # Filtering for values between q25-5*iqr and q75+5*iqr
            df = df.query('(@q25 - 5 * @iqr) <= {} <= (@q75 + 5 * @iqr)'.format(col))

        # Convert the unit of expense-related & loan-related variables from Rupee to USD
        conv = 9.1768
        for col in ['spandana_amt_1', 'othermfi_amt_1', 'bank_amt_1',
                    'informal_amt_1', 'durables_exp_mo_pc_1', 
                    'nondurable_exp_mo_pc_1', 'food_exp_mo_pc_1',
                    'health_exp_mo_pc_1', 'temptation_exp_mo_pc_1',
                    'festival_exp_mo_pc_1']:
            df[col] = df[col].div(conv)

    return df


def load_ifmr_microfinance(
    data_path=None,
    load_raw_data=False, # = True loads a raw csv from the Stata .dta file
    download_if_missing=True
):
    """
    Parameters
    ----------
    load_raw_data : loads an unformated version of the data (detault=False)
    data_path : str, optional (default=None)
        - Specify another download and cache folder for the dataset.
        - By default the dataset will be stored in the 'datasets' folder in the cwd.
    download_if_missing : bool, optional (default=True)
        - Download the dataset if it is not downloaded before using 'download_ifmr_microfinance'.
    
    Returns
    -------
    dataset : dict object with the following attributes:
    dataset.description : str
        - A description of the IFMR microfinance data.
    dataset.dataset_full : ndarray, shape (5328, 61)
        - The full dataset with features, treatment, and target variables
    dataset.data : ndarray, shape (5328, 58)
        - Each row corresponding to the 58 feature values in order (note that other target can be a feature).
    dataset.feature_names : list, size 58
        - List of feature names.
    dataset.treatment : ndarray, shape (5328,)
        - Each value corresponds to the treatment (1 = treat, 0 = control).
    dataset.target_biz_index : numpy array of shape (5328,)
        - Each value corresponds to the business index of each of the participants.
    dataset.target_women_emp : numpy array of shape (5328,)
        - Each value corresponds to the women's empowerment index of each of the participants.
    """
    # Check that the dataset exists
    data_path, dataset_path = get_download_paths(data_path, 'datasets', 'ifmr_microfinance')
    if not os.path.exists(dataset_path):
        if download_if_missing:
            downlaod_ifmr_microfinance(data_path)
        else:
            raise FileNotFoundError(
                "The dataset does not exist."
                "Use the 'downlaod_ifmr_microfinance' function to download the dataset."
            )

    # Load formated or raw data
    if not load_raw_data:
        df = __format_data(dataset_path)
    else:
        df = __format_data(dataset_path, load_raw_data=True)

    description = "The data comes from The Centre for Micro Finance at the Institute for Financial Management Research (IFMR) (Chennai, India)"\
                  "The feature set can be used to derive the effects of microfinance on various post-treatment indexes."\
                  "Specifically we will focus on the post-treatment business and women's empowerment indexes."\
                  "The other target value can be added into the dataset as a feature."

    # Fields dropped to split the data for the user
    drop_fields = ['biz_index_all_1', 'women_emp_index_1', 'treatment']
    
    data = {
        'description': description,
        'dataset_full' : df.values,
        'data': df.drop(drop_fields, axis=1).values,
        'feature_names': np.array(list(filter(lambda x: x not in drop_fields, df.columns))),
        'treatment': df['treatment'].values,
        # The target that isn't of interest can also be used as a feature
        'target_biz_index': df['biz_index_all_1'].values,
        'target_women_emp': df['women_emp_index_1'].values
    }

    return data


class GRF: # GeneralizedRandomForest

    def grf_fit(self, X, y, w):
        """
        Trains a model using the Generalized Random Forest algorithm

        Based on
        --------
        - "Generalized Random Forests" (Athey et al, 2018) 
        - The accompanything R package: https://github.com/grf-labs/grf/tree/master/r-package/grf/
        - grf documentation: https://grf-labs.github.io/grf/
        - "Estimation and Inference of Heterogeneous Treatment Effects using Random Forests" (Wager and Athey 2018)

        Requirements
        ------------
        - pandas : used for grouping via the DataFrame module
        - scikit-learn : used for training via sklearn.module.model_class.fit()
        - For model options see : https://scikit-learn.org/stable/supervised_learning.html#supervised-learning

        Parameters
        ----------
        X : dataframe of covariates (type(s): int, float)
        y : vector of unit reponses (type: int, float)
        w : binary vector designating the original treatment group allocation across units (type: float)
        model_class : the class of supervised learning model to use (base: LinearRegression)

        Returns
        -------
        - A trained model
        """
        model = 1
        return model

    def grf_pred(self, model, X_pred, continuous = False):
        """
        Makes predicitons using the Generalized Random Forest algorithm

        Based on
        --------
        - "Generalized Random Forests" (Athey et al, 2018) 
        - The accompanything R package: https://github.com/grf-lab∂s/grf/tree/master/r-package/grf/
        - grf documentation: https://grf-labs.github.io/grf/
        - "Estimation and Inference of Heterogeneous Treatment Effects using Random Forests" (Wager and Athey 2018)

        Requirements
        ------------
        - NumPy : for arrays
        - scikit-learn : used for predictions

        Parameters
        ----------
        model : a model that has been fit using the Generalized Random Forest algorithm
        X_pred : new data on which to make a prediction

        Returns
        -------
        - A NumPy array of predicted outcomes for each unit in X_pred based on treatment assignment
        """
        pred_tuples = 1
        return np.array(pred_tuples)


"""
A dataset on medical trials to combat primary biliary cirrhosis of the liver from the Mayo Clinic

Description found at
--------------------
https://www.mayo.edu/research/documents/pbchtml/DOC-10027635

A similar dataset can be found in appendix D of
-----------------------------------------------
Counting Processes and Survival Analysis, Thomas R. Fleming David P. Harrington, Wiley, Septemher 2005
https://onlinelibrary.wiley.com/doi/book/10.1002/9781118150672
"""
  
import os
import numpy as np
import pandas as pd
from causeinfer.datasets.download_utils import download_file, get_download_paths

def download_mayo_pbc(
    data_path=None,
    url='http://www.mayo.edu/research/documents/pbcdat/DOC-10026921'
):
    """
    Downloads the dataset from the Mayo Clinic's research documents.

    Result
    ------
    - The text file 'mayo_pbc' in a 'datasets' folder, unless otherwise specified
    """
    data_path, dataset_path = get_download_paths(data_path, 'datasets', 'mayo_pbc.text')
    if not os.path.isdir(data_path):
        os.makedirs(data_path)

    if not os.path.exists(dataset_path):
        download_file(url, dataset_path, zip_file = False)


def __format_data(dataset_path):
    """
    Formats the data upon loading for consistent data preparation.

    The original file is a text file with inconsistent spacing, and periods for NaNs. Furthermore, process only loads 
    those units that took part in the randomized trial, as there are 106 cases that were monitored, but not in the trial.
    """
    # Read in the text file
    with open(dataset_path, 'r') as f:
        data = f.read().splitlines()
    
    # The following converts the text file into a list of lists
    # The three iterations account for initial spaces for single, double, and tripple digit numbers
    data_list_of_lists = [data[i][2:].replace('    ', ',').replace('   ', ',').replace('  ', ',').replace(' ', ',').split(',')[1:] for i in range(10)]
    data_list_of_lists.extend([data[i][1:].replace('    ', ',').replace('   ', ',').replace('  ', ',').replace(' ', ',').split(',')[1:] for i in list(range(99))[10:]])
    data_list_of_lists.extend([data[i].replace('    ', ',').replace('   ', ',').replace('  ', ',').replace(' ', ',').split(',')[1:] for i in list(range(312))[99:]])

    df = pd.DataFrame(data_list_of_lists)

    col_names = ['days_since_register', 'status', 'treatment', 'age',
                 'sex', 'ascites', 'hepatomegaly', 'spiders', 'edema',
                 'bilirubin', 'cholesterol', 'albumin', 'copper', 
                 'alkaline', 'SGOT', 'triglicerides', 'platelets', 
                 'prothrombin', 'histologic stage']
    df.columns = col_names

    # Filling NaNs with column averages (they occur in cholesterol, copper, triglicerides and platelets)
    df = df.replace('.', np.nan)
    df = df.astype(float)
    df.fillna(df.mean(), inplace=True)

    # Replace control from 2 to 0
    df.loc[df['treatment'] == 2, 'treatment'] = 0

    return df


def load_mayo_pbc(
    data_path=None,
    # load_raw_data=False,
    download_if_missing=True
):
    """
    Parameters
    ----------
    load_raw_data : - not included, as original data isn't in table form -
    data_path : str, optional (default=None)
        - Specify another download and cache folder for the dataset.
        - By default the dataset will be stored in the 'datasets' folder in the cwd.
    download_if_missing : bool, optional (default=True)
        - Download the dataset if it is not downloaded before using 'download_mayo_pbc'.
    
    Returns
    -------
    dataset : dict object with the following attributes:
    dataset.description : str
        - A description of the Mayo Clinic PBC dataset.
    dataset.dataset_full : ndarray, shape (312, 19)
        - The full dataset with features, treatment, and target variables
    dataset.data : ndarray, shape (312, 17)
        - Each row corresponding to the 8 feature values in order.
    dataset.feature_names : list, size 17
        - List of feature names.
    dataset.treatment : ndarray, shape (312,)
        - Each value corresponds to the treatment (1 = treat, 0 = control).
    dataset.target : numpy array of shape (312,)
        - Each value corresponds to one of the outcomes (0 = alive, 1 = liver transplant, 2 = dead).
    """
    # Check that the dataset exists
    data_path, dataset_path = get_download_paths(data_path, 'datasets', 'mayo_pbc.text')
    if not os.path.exists(dataset_path):
        if download_if_missing:
            download_mayo_pbc(data_path)
        else:
            raise FileNotFoundError(
                "The dataset does not exist."
                "Use the 'download_mayo_pbc' function to download the dataset."
            )

    # Load formated data
    df = __format_data(dataset_path)

    description = 'The data is from the Mayo Clinic trial in primary biliary cirrhosis (PBC) of the liver conducted between 1974 and 1984.' \
                  'A total of 424 PBC patients, referred to Mayo Clinic during that ten-year interval, met eligibility criteria for the randomized placebo controlled trial of the drug D-penicillamine.' \
                  'The first 312 cases in the data set participated in the randomized trial and contain largely complete data.'\
                  'For modeling purposes, alive (target=0) will be modelled against a resulting transplant (1) and death (2).'

    # Fields dropped to split the data for the user
    drop_fields = ['status', 'treatment']
    
    data = {
        'description': description,
        'dataset_full' : df.values,
        'data': df.drop(drop_fields, axis=1).values,
        'feature_names': np.array(list(filter(lambda x: x not in drop_fields, df.columns))),
        'treatment': df['treatment'].values,
        'target': df['status'].values
    }

    return data


"""
An email marketing dataset from Kevin Hillstrom's MineThatData blog

Description found at
--------------------
https://blog.minethatdata.com/2008/03/minethatdata-e-mail-analytics-and-data.html
"""
  
import os
import numpy as np
import pandas as pd
from causeinfer.datasets.download_utils import download_file, get_download_paths

def download_hillstrom(
    data_path=None,
    url='http://www.minethatdata.com/Kevin_Hillstrom_MineThatData_E-MailAnalytics_DataMiningChallenge_2008.03.20.csv'
):
    """
    Downloads the dataset from Kevin Hillstrom's blog

    Result
    ------
    - The data 'hillstrom.csv' in a 'datasets' folder, unless otherwise specified
    """
    data_path, dataset_path = get_download_paths(data_path, 'datasets', 'hillstrom.csv')
    if not os.path.isdir(data_path):
        os.makedirs(data_path)

    if not os.path.exists(dataset_path):
        download_file(url, dataset_path, zip_file = False)


def __format_data(df):
    """
    Formats the data upon loading for consistent data preparation
    """
    # Split away the history segment index
    df['history_segment'] = df['history_segment'].apply(lambda s: s.split(') ')[1])
    
    # Create dummy columns for zip_code, history_segment, and channel
    dummy_cols = ['zip_code', 'history_segment', 'channel']
    for col in dummy_cols:
        df = pd.get_dummies(df, columns=[col], prefix=col)

    # Encode the segment column
    segment_encoder = {'No E-Mail': 0, 'Mens E-Mail': 1, 'Womens E-Mail': 2}
    df['segment'] = df['segment'].apply(lambda x: segment_encoder[x])
    return df


def load_hillstrom(
    data_path=None,
    load_raw_data=False,
    download_if_missing=True
):
    """
    Parameters
    ----------
    load_raw_data : bool, default: False
        - Indicates whether the raw data should be loaded without '__format_data'.
    data_path : str, optional (default=None)
        - Specify another download and cache folder for the dataset.
        - By default the dataset will be stored in the 'datasets' folder in the cwd.
    download_if_missing : bool, optional (default=True)
        - Download the dataset if it is not downloaded before using 'download_hillstrom'.
    
    Returns
    -------
    dataset : dict object with the following attributes:
    dataset.description : str
        - A description of the Hillstrom email marketing dataset.
    dataset.dataset_full : ndarray, shape (64000, 12)
        - The full dataset with features, treatment, and target variables
    dataset.data : ndarray, shape (64000, 8)
        - Each row corresponding to the 8 feature values in order.
    dataset.feature_names : list, size 8
        - List of feature names.
    dataset.treatment : ndarray, shape (64000,)
        - Each value corresponds to the treatment.
    dataset.target : numpy array of shape (64000,)
        - Each value corresponds to one of the outcomes. By default, it's `spend` outcome (look at `target_spend` below).
    dataset.target_spend : numpy array of shape (64000,)
        - Each value corresponds to how much customers spent during the two-week outcome period.
    dataset.target_visit : numpy array of shape (64000,)
        - Each value corresponds to whether people visited the site during the two-week outcome period.
    dataset.target_conversion : numpy array of shape (64000,)
        - Each value corresponds to whether they purchased at the site (i.e. converted) during the two-week outcome period.
    """
    # Check that the dataset exists
    data_path, dataset_path = get_download_paths(data_path, 'datasets', 'hillstrom.csv')
    if not os.path.exists(dataset_path):
        if download_if_missing:
            download_hillstrom(data_path)
        else:
            raise FileNotFoundError(
                "The dataset does not exist."
                "Use the 'download_hillstrom' function to download the dataset."
            )

    # Load formated or raw data
    df = pd.read_csv(dataset_path)
    if not load_raw_data:
        df = __format_data(df)

    description = 'The Hilstrom dataset contains 64,000 customers who purchased within twelve months.' \
                  'The customers were involved in an e-mail marketing test.' \
                  '1/3 were randomly chosen to receive an e-mail campaign featuring Mens merchandise.' \
                  '1/3 were randomly chosen to receive an e-mail campaign featuring Womens merchandise.' \
                  '1/3 were randomly chosen to not receive an e-mail campaign.' \
                  'During a period of two weeks following the e-mail campaign, results were tracked.' \
                  'Targeting for causal inference can be derived using visit, conversion, or total spent.'

    # Fields dropped to split the data for the user
    drop_fields = ['spend', 'visit', 'conversion', 'segment']
    
    data = {
        'description': description,
        'dataset_full' : df.values,
        'data': df.drop(drop_fields, axis=1).values,
        'feature_names': np.array(list(filter(lambda x: x not in drop_fields, df.columns))),
        'treatment': df['segment'].values,
        'target': df['spend'].values,
        'target_spend': df['spend'].values,
        'target_visit': df['visit'].values,
        'target_conversion': df['conversion'].values,
    }

    return data


import os
import requests
import urllib
import zipfile

def download_file(url: str, output_path: str, zip_file = False):
    """
    Downloads a file from a url to a specified path

    Parameters
    ----------
    url : the URL from which the file can be downloaded from
    output_path : a user specified path, which defaults to a 'files' folder in the cwd
    """
    # Check if the file exists, and delete if so
    if os.path.isfile(output_path):
        os.remove(output_path)

    print("Downloading file to '{}' ...".format(output_path))
    
    res = requests.get(url)
    # Check if the response is ok (200)
    status_code = int(res.status_code)
    if status_code == 200:
        if zip_file == True:
            file = urllib.request.urlretrieve(url, "{}.zip".format(output_path))
            with zipfile.ZipFile("{}.zip".format(output_path), 'r') as zip_ref:
                print("Unzipping '{}'.zip ...".format(output_path))
                zip_ref.extractall("{}.zip".format(output_path).split(".zip")[0])
                os.remove("{}.zip".format(output_path))
                print("File unzipped - deleting .zip file")

        else:
            with open(output_path, 'wb') as file:
                # A chunk of 128 bytes
                for chunk in res:
                    file.write(chunk)
                    
    elif status_code == 404:
        raise Exception('Wrong URL (' + url + ').')


def get_download_paths(file_path, file_folder = 'files', file_name = 'file'):
    """
    Derives paths for a file folder and a file path
    """
    if file_path is None:
        file_path = os.path.join(os.getcwd() + '/' + file_folder)

    file_folder = os.path.join(file_path + '/' + file_name)
    return file_path, file_folder


  
"""
- The module contains causeinfer's algorithms

Inlcuded algorithmic methods:
1. Two Model Approach
2. Interaction Term Approach - Lo 2002
3. Response Transformation Approach - Lai 2006; Kane, Lo and Zheng 2014
4. Generalized Random Forest - Athey, Tibshirani, and Wager 2018
"""

from .two_model import TwoModel
from .interaction_term import InteractionTerm
from .response_transformation import ResponseTransformation
from .generalized_random_forest import GRF

from .evaluation import Evaluation


class CauseInfer:

    # ----------------------------------------------------------------------------------------------------------------------
    # 1. Two Model Approach
    # ----------------------------------------------------------------------------------------------------------------------


    def two_model_fit(X, y, w, module = "linear_model", model_class = "LinearRegression"):
        """
        Trains a model using the "Two Model Approach" (Double Model, Separate Model)

        Based on
        --------
        - "A Literature Survey and Experimental Evaluation of the State-of-the-Art in Uplift Modeling:
        A Stepping Stone Toward the Development of Prescriptive Analytics" (Devriendt, 2018) 
        - "Incremental Value Modeling" (Hansotia, 2002)

        Requirements
        ------------
        - pandas : used for grouping via the DataFrame module
        - scikit-learn : used for training via sklearn.module.model_class.fit()
        - For model options see : https://scikit-learn.org/stable/supervised_learning.html#supervised-learning

        Parameters
        ----------
        X : dataframe of covariates (type(s): int, float)
        y : vector of unit reponses (type: int, float)
        w : binary vector designating the original treatment group allocation across units (type: float)
        model_class : the class of supervised learning model to use (base: LinearRegression)

        Returns
        -------
        - Two trained models - one for the training group, and one for control
        """
        from sklearn import linear_model
        from sklearn import ensemble
        import sklearn
        import pandas as pd 

        df = pd.DataFrame(X, w)

        model_trn = sklearn.module.model_class.fit(X = df[df[w==1]], y=y)   # fit for treatment group
        model_ctrl = sklearn.module.model_class.fit(X = df[df[w==0]], y=y)  # fit for control group

        res = list(model_treatment = model_trn,
                model_control = model_ctrl, 
                model_class = model_class)

        return(res)

    def two_model_pred(models, X_pred, continuous = False):
        """
        Makes predicitons using the "Two Model Approach" (Double Model, Separate Model)

        Based on
        --------
        - "A Literature Survey and Experimental Evaluation of the State-of-the-Art in Uplift Modeling:
        A Stepping Stone Toward the Development of Prescriptive Analytics" (Devriendt, 2018) 
        - "Incremental Value Modeling" (Hansotia, 2002)

        Requirements
        ------------
        - NumPy : for arrays
        - scikit-learn : used for predictions

        Parameters
        ----------
        models : a list of two models that have been fit on the treatment and control groups respectively
        X_pred : new data on which to make a prediction

        Returns
        -------
        - A NumPy array of predicted outcomes for each unit in X_pred based on treatment assignment
        """
        import numpy as np

        pred_trt = models[model_treatment].predict(X_pred)
        pred_ctrl = models[model_control].predict(X_pred)

        tuples = [(pred_trt[i], pred_ctrl(i)) for i in list(range(len(X_pred)))]

        return np.array(tuples)


    # ----------------------------------------------------------------------------------------------------------------------
    # 2. Interaction Term Approach - Lo
    # ----------------------------------------------------------------------------------------------------------------------


    def interaction_term_fit(X, y, w, module = "linear_model", model_class = "LinearRegression"):
        """
        Trains a model using the "Interaction Term Approach" (Dummy Treatment Approach)

        Based on
        --------
        - "A Literature Survey and Experimental Evaluation of the State-of-the-Art in Uplift Modeling:
        A Stepping Stone Toward the Development of Prescriptive Analytics" (Devriendt, 2018) 
        - "The True Lift Model" (Lo, 2002)

        Requirements
        ------------
        - pandas : used for grouping via the DataFrame module
        - scikit-learn : used for training via sklearn.module.model_class.fit()
        - For model options see : https://scikit-learn.org/stable/supervised_learning.html#supervised-learning

        Parameters
        ----------
        X : dataframe of covariates (type(s): int, float)
        y : vector of unit reponses (type: int, float)
        w : binary vector designating the original treatment group allocation across units (type: float)
        model_class : the class of supervised learning model to use (base: LinearRegression)

        Returns
        -------
        - A trained model
        """
        from sklearn import linear_model
        from sklearn import ensemble
        import sklearn
        import pandas as pd 

        # create interaction terms
        xT = X * w

        # new data now includes the interaction term
        df = pd.DataFrame(X, w, xT)

        model = sklearn.module.model_class.fit(X = df, y=y)
        
        return model

    def interaction_term_pred(model, X_pred, y_id = "y", w_id ="w", continuous = False):
        """
        Makes predicitons using the "Interaction Term Approach" (Dummy Treatment Approach)

        Based on
        --------
        - "A Literature Survey and Experimental Evaluation of the State-of-the-Art in Uplift Modeling:
        A Stepping Stone Toward the Development of Prescriptive Analytics" (Devriendt, 2018) 
        - "The True Lift Model" (Lo, 2002)

        Requirements
        ------------
        - NumPy : for arrays
        - scikit-learn : used for predictions

        Parameters
        ----------
        models : a model that has been fit using the "Interaction Term Approach"
        X_pred : new data on which to make a prediction

        Returns
        -------
        - A NumPy array of predicted outcomes for each unit in X_pred based on treatment assignment
        """
        import numpy as np
        import pandas as pd
    
        predictors = names(X_pred)[(names(X_pred) != y_id) & (names(X_pred) != w_id)]
        
        xt_trt = X_pred["predictors"] * 1
        colnames(xt_trt) = paste("Int", colnames(xt_trt), sep = "_")
        df_trt = pd.DataFrame(X_pred, w=1, xt_trt)  
        
        xt_ctrl = X_pred["predictors"] * 0
        colnames(xt_ctrl) = paste("Int", colnames(xt_ctrl), sep = "_")
        df_ctrl = pd.DataFrame(X_pred, w=0, xt_ctrl)
        
        pred_trt = model.predict(df_trt)
        pred_ctrl = model.predict(df_ctrl)

        pred_tuples = [(pred_trt[i], pred_ctrl(i)) for i in list(range(len(X_pred)))]

        return np.array(pred_tuples)


    # ----------------------------------------------------------------------------------------------------------------------
    # 3. Response Transformation Approach - Lai, Kane
    # ----------------------------------------------------------------------------------------------------------------------


    def response_transformation_fit(X, y, w, module = "linear_model", model_class = "LinearRegression"):
        """
        Trains a model using the "Response Treatment Approach" (Influential Marketing)

        Based on
        --------
        - "A Literature Survey and Experimental Evaluation of the State-of-the-Art in Uplift Modeling:
        A Stepping Stone Toward the Development of Prescriptive Analytics" (Devriendt, 2018) 
        - "Influential Marketing: A New Direct Marketing Strategy Addressing the Existence of Voluntary Buyers" (Lai, 2006)
        - "Mining for the truly responsive customers and prospects using true-lift modeling: 
        Comparison of new and existing methods" (Kane, 2014)

        Requirements
        ------------
        - pandas : used for grouping via the DataFrame module
        - scikit-learn : used for training via sklearn.module.model_class.fit()
        - For model options see : https://scikit-learn.org/stable/supervised_learning.html#supervised-learning

        Parameters
        ----------
        X : dataframe of covariates (type(s): int, float)
        y : vector of unit reponses (type: int, float)
        w : binary vector designating the original treatment group allocation across units (type: float)
        model_class : the class of supervised learning model to use (base: LinearRegression)

        Returns
        -------
        - A trained model
        """
        from sklearn import linear_model
        from sklearn import ensemble
        import sklearn
        import pandas as pd 

        df = pd.DataFrame(X, w_y = np.nan)

        df["w_y"][y == 1 & w == 1] = "TR" # Treated responders
        df["w_y"][y == 0 & w == 1] = "TN" # Treated non-responders
        df["w_y"][y == 1 & w == 0] = "CR" # Control responders
        df["w_y"][y == 0 & w == 0] = "CN" # Control non-responders

        model = sklearn.module.model_class.fit(X = df, y=y)
        
        return model

    def response_transformation_pred(model, X_pred, y_id = "y", w_id ="w", generalized = True, continuous = False):
        """
        Makes predicitons using the "Response Treatment Approach" (Influential Marketing)

        Based on
        --------
        - "A Literature Survey and Experimental Evaluation of the State-of-the-Art in Uplift Modeling:
        A Stepping Stone Toward the Development of Prescriptive Analytics" (Devriendt, 2018) 
        - "Influential Marketing: A New Direct Marketing Strategy Addressing the Existence of Voluntary Buyers" (Lai, 2006)
        - "Mining for the truly responsive customers and prospects using true-lift modeling: 
        Comparison of new and existing methods" (Kane, 2014)

        Requirements
        ------------
        - NumPy : for arrays
        - scikit-learn : used for predictions

        Parameters
        ----------
        model : a model that has been fit using the "Response Treatment Approach"
        X_pred : new data on which to make a prediction

        Returns
        -------
        - A NumPy array of predicted outcomes for each unit in X_pred based on treatment assignment
        """
        prob_C = prop.table(table(X_pred[str(w_id)]))[0] # Percent of units that are control
        prob_T = prop.table(table(X_pred[str(w_id)]))[1] # Percent of units that are treatment

        X_pred["w_y"] <- np.nan
        # The following splits the units into known classes, with the goal then being the derive those charactaristics for
        # Persuadables and Do Not Disturbs - the separated positive and negative classes - from the neutral classes 
        X_pred["w_y"][X_pred["y_id"] == 1 & X_pred["w_id"] == 1] = "TR" # Treated responders (Business: Sure Things or Persuadables)
        X_pred["w_y"][X_pred["y_id"] == 0 & X_pred["w_id"] == 1] = "TN" # Treated non-responders (Business: Lost Causes or Do Not Disturbs)
        X_pred["w_y"][X_pred["y_id"] == 1 & X_pred["w_id"] == 0] = "CR" # Control responders (Business: Sure Things or Do Not Disturbs)
        X_pred["w_y"][X_pred["y_id"] == 0 & X_pred["w_id"] == 0] = "CN" # Control non-responders (Business: Lost Causes or Persuadables)

        pred <- model.predict(X_pred)

        if generalized:
            # The generalized approach as suggested in Kane 2014
            pr_y1_w1 = ((pred["TR"] / prob_T) + (pred["CN"] / prob_C))
            pr_y1_w0 = ((pred["TN"] / prob_T) + (pred["CR"] / prob_C))
        else:
            # The original approach as suggested in Lai 2006 with the inclusion of Do Not Disturb units
            pr_y1_w1 = pred["TR"] + pred["CN"]
            pr_y1_w0 = pred["TN"] + pred["CR"]
        
        pred_tuples = [(pr_y1_w1[i], pr_y1_w0(i)) for i in list(range(len(X_pred)))]

        return np.array(pred_tuples)


    # ----------------------------------------------------------------------------------------------------------------------
    # 4. Generalized Random Forest (GRF) - Athey, Wager
    # ----------------------------------------------------------------------------------------------------------------------


    def grf_pred(X, y, w):
        """
        Trains a model using the Generalized Random Forest algorithm

        Based on
        --------
        - "Generalized Random Forests" (Athey et al, 2018) 
        - The accompanything R package: https://github.com/grf-labs/grf/tree/master/r-package/grf/
        - grf documentation: https://grf-labs.github.io/grf/
        - "Estimation and Inference of Heterogeneous Treatment Effects using Random Forests" (Wager and Athey 2018)

        Requirements
        ------------
        - pandas : used for grouping via the DataFrame module
        - scikit-learn : used for training via sklearn.module.model_class.fit()
        - For model options see : https://scikit-learn.org/stable/supervised_learning.html#supervised-learning

        Parameters
        ----------
        X : dataframe of covariates (type(s): int, float)
        y : vector of unit reponses (type: int, float)
        w : binary vector designating the original treatment group allocation across units (type: float)
        model_class : the class of supervised learning model to use (base: LinearRegression)

        Returns
        -------
        - A trained model
        """
        model = 1
        return model

    def grf_pred(model, X_pred, continuous = False):
        """
        Makes predicitons using the Generalized Random Forest algorithm

        Based on
        --------
        - "Generalized Random Forests" (Athey et al, 2018) 
        - The accompanything R package: https://github.com/grf-lab∂s/grf/tree/master/r-package/grf/
        - grf documentation: https://grf-labs.github.io/grf/
        - "Estimation and Inference of Heterogeneous Treatment Effects using Random Forests" (Wager and Athey 2018)

        Requirements
        ------------
        - NumPy : for arrays
        - scikit-learn : used for predictions

        Parameters
        ----------
        model : a model that has been fit using the Generalized Random Forest algorithm
        X_pred : new data on which to make a prediction

        Returns
        -------
        - A NumPy array of predicted outcomes for each unit in X_pred based on treatment assignment
        """
        pred_tuples = 1
        return np.array(pred_tuples)


    # ----------------------------------------------------------------------------------------------------------------------
    # 5. Evaluation Metrics
    # ----------------------------------------------------------------------------------------------------------------------


    def generalized_random_forest_confidence_intervals(): # import as grf_ci

        return GRF_CIs

    def qini_AUUC_scores(qini = True):

        return qini_AUUC


class InteractionTerm:
    
    def interaction_term_fit(self, X, y, w, module = "linear_model", model_class = "LinearRegression"):
        """
        Trains a model using the "Interaction Term Approach" (Dummy Treatment Approach)

        Based on
        --------
        - "A Literature Survey and Experimental Evaluation of the State-of-the-Art in Uplift Modeling:
        A Stepping Stone Toward the Development of Prescriptive Analytics" (Devriendt, 2018) 
        - "The True Lift Model" (Lo, 2002)

        Requirements
        ------------
        - pandas : used for grouping via the DataFrame module
        - scikit-learn : used for training via sklearn.module.model_class.fit()
        - For model options see : https://scikit-learn.org/stable/supervised_learning.html#supervised-learning

        Parameters
        ----------
        X : dataframe of covariates (type(s): int, float)
        y : vector of unit reponses (type: int, float)
        w : binary vector designating the original treatment group allocation across units (type: float)
        model_class : the class of supervised learning model to use (base: LinearRegression)

        Returns
        -------
        - A trained model
        """
        from sklearn import linear_model
        from sklearn import ensemble
        import sklearn
        import pandas as pd 

        # create interaction terms
        xT = X * w

        # new data now includes the interaction term
        df = pd.DataFrame(X, w, xT)

        model = sklearn.module.model_class.fit(X = df, y=y)
        
        return model

    def interaction_term_pred(self, model, X_pred, y_id = "y", w_id ="w", continuous = False):
        """
        Makes predicitons using the "Interaction Term Approach" (Dummy Treatment Approach)

        Based on
        --------
        - "A Literature Survey and Experimental Evaluation of the State-of-the-Art in Uplift Modeling:
        A Stepping Stone Toward the Development of Prescriptive Analytics" (Devriendt, 2018) 
        - "The True Lift Model" (Lo, 2002)

        Requirements
        ------------
        - NumPy : for arrays
        - scikit-learn : used for predictions

        Parameters
        ----------
        models : a model that has been fit using the "Interaction Term Approach"
        X_pred : new data on which to make a prediction

        Returns
        -------
        - A NumPy array of predicted outcomes for each unit in X_pred based on treatment assignment
        """
        import numpy as np
        import pandas as pd
    
        predictors = names(X_pred)[(names(X_pred) != y_id) & (names(X_pred) != w_id)]
        
        xt_trt = X_pred["predictors"] * 1
        colnames(xt_trt) = paste("Int", colnames(xt_trt), sep = "_")
        df_trt = pd.DataFrame(X_pred, w=1, xt_trt)  
        
        xt_ctrl = X_pred["predictors"] * 0
        colnames(xt_ctrl) = paste("Int", colnames(xt_ctrl), sep = "_")
        df_ctrl = pd.DataFrame(X_pred, w=0, xt_ctrl)
        
        pred_trt = model.predict(df_trt)
        pred_ctrl = model.predict(df_ctrl)

        pred_tuples = [(pred_trt[i], pred_ctrl(i)) for i in list(range(len(X_pred)))]

        return np.array(pred_tuples)


class ResponseTransformation:

    def response_transformation(self, X, y, w, module = "linear_model", model_class = "LinearRegression"):
        """
        Trains a model using the "Response Transformation Approach" (Influential Marketing)

        Based on
        --------
        - "A Literature Survey and Experimental Evaluation of the State-of-the-Art in Uplift Modeling:
        A Stepping Stone Toward the Development of Prescriptive Analytics" (Devriendt, 2018) 
        - "Influential Marketing: A New Direct Marketing Strategy Addressing the Existence of Voluntary Buyers" (Lai, 2006)
        - "Mining for the truly responsive customers and prospects using true-lift modeling: 
        Comparison of new and existing methods" (Kane, 2014)

        Requirements
        ------------
        - pandas : used for grouping via the DataFrame module
        - scikit-learn : used for training via sklearn.module.model_class.fit()
        - For model options see : https://scikit-learn.org/stable/supervised_learning.html#supervised-learning

        Parameters
        ----------
        X : dataframe of covariates (type(s): int, float)
        y : vector of unit reponses (type: int, float)
        w : binary vector designating the original treatment group allocation across units (type: float)
        model_class : the class of supervised learning model to use (base: LinearRegression)

        Returns
        -------
        - A trained model
        """
        from sklearn import linear_model
        from sklearn import ensemble
        import sklearn
        import pandas as pd 

        df = pd.DataFrame(X, w_y = np.nan)

        df["w_y"][y == 1 & w == 1] = "TR" # Treated responders
        df["w_y"][y == 0 & w == 1] = "TN" # Treated non-responders
        df["w_y"][y == 1 & w == 0] = "CR" # Control responders
        df["w_y"][y == 0 & w == 0] = "CN" # Control non-responders

        model = sklearn.module.model_class.fit(X = df, y=y)
        
        return model

    def response_transformation_pred(self, model, X_pred, y_id = "y", w_id ="w", generalized = True, continuous = False):
        """
        Makes predicitons using the "Response Treatment Approach" (Influential Marketing)

        Based on
        --------
        - "A Literature Survey and Experimental Evaluation of the State-of-the-Art in Uplift Modeling:
        A Stepping Stone Toward the Development of Prescriptive Analytics" (Devriendt, 2018) 
        - "Influential Marketing: A New Direct Marketing Strategy Addressing the Existence of Voluntary Buyers" (Lai, 2006)
        - "Mining for the truly responsive customers and prospects using true-lift modeling: 
        Comparison of new and existing methods" (Kane, 2014)

        Requirements
        ------------
        - NumPy : for arrays
        - scikit-learn : used for predictions

        Parameters
        ----------
        model : a model that has been fit using the "Response Treatment Approach"
        X_pred : new data on which to make a prediction

        Returns
        -------
        - A NumPy array of predicted outcomes for each unit in X_pred based on treatment assignment
        """
        prob_C = prop.table(table(X_pred[str(w_id)]))[0] # Percent of units that are control
        prob_T = prop.table(table(X_pred[str(w_id)]))[1] # Percent of units that are treatment

        X_pred["w_y"] <- np.nan
        # The following splits the units into known classes, with the goal then being the derive those charactaristics for
        # Persuadables and Do Not Disturbs - the separated positive and negative classes - from the neutral classes 
        X_pred["w_y"][X_pred["y_id"] == 1 & X_pred["w_id"] == 1] = "TR" # Treated responders (Business: Sure Things or Persuadables)
        X_pred["w_y"][X_pred["y_id"] == 0 & X_pred["w_id"] == 1] = "TN" # Treated non-responders (Business: Lost Causes or Do Not Disturbs)
        X_pred["w_y"][X_pred["y_id"] == 1 & X_pred["w_id"] == 0] = "CR" # Control responders (Business: Sure Things or Do Not Disturbs)
        X_pred["w_y"][X_pred["y_id"] == 0 & X_pred["w_id"] == 0] = "CN" # Control non-responders (Business: Lost Causes or Persuadables)

        pred <- model.predict(X_pred)

        if generalized:
            # The generalized approach as suggested in Kane 2014
            pr_y1_w1 = ((pred["TR"] / prob_T) + (pred["CN"] / prob_C))
            pr_y1_w0 = ((pred["TN"] / prob_T) + (pred["CR"] / prob_C))
        else:
            # The original approach as suggested in Lai 2006 with the inclusion of Do Not Disturb units
            pr_y1_w1 = pred["TR"] + pred["CN"]
            pr_y1_w0 = pred["TN"] + pred["CR"]
        
        pred_tuples = [(pr_y1_w1[i], pr_y1_w0(i)) for i in list(range(len(X_pred)))]

        return np.array(pred_tuples)


from setuptools import setup, find_namespace_packages

with open("README.md", "r") as fh:
    long_description = fh.read()

setup_args = dict(
    name='causeinfer',
    version='0.0.2',
    description='Causal inference/uplift in Python',
    long_description=long_description,
    long_description_content_type="text/markdown",
    packages=find_namespace_packages(),
    license='MIT',
    url="https://github.com/andrewtavis/causeinfer",
    author='Andrew Tavis McAllister',
    author_email='andrew.t.mcallister@gmail.com'
)

install_requires = [
    'numpy',
    'pandas',
    'scikit-learn'
]

if __name__ == '__main__':
    setup(**setup_args, install_requires=install_requires)


class TwoModel:

    def two_mode_fitl(self, X, y, w, module = "linear_model", model_class = "LinearRegression"):
        """
        Trains a model using the "Two Model Approach" (Double Model, Separate Model)

        Based on
        --------
        - "A Literature Survey and Experimental Evaluation of the State-of-the-Art in Uplift Modeling:
        A Stepping Stone Toward the Development of Prescriptive Analytics" (Devriendt, 2018) 
        - "Incremental Value Modeling" (Hansotia, 2002)

        Requirements
        ------------
        - pandas : used for grouping via the DataFrame module
        - scikit-learn : used for training via sklearn.module.model_class.fit()
        - For model options see : https://scikit-learn.org/stable/supervised_learning.html#supervised-learning

        Parameters
        ----------
        X : dataframe of covariates (type(s): int, float)
        y : vector of unit reponses (type: int, float)
        w : binary vector designating the original treatment group allocation across units (type: float)
        model_class : the class of supervised learning model to use (base: LinearRegression)

        Returns
        -------
        - Two trained models - one for the training group, and one for control
        """
        from sklearn import linear_model
        from sklearn import ensemble
        import sklearn
        import pandas as pd 

        df = pd.DataFrame(X, w)

        model_trn = sklearn.module.model_class.fit(X = df[df[w==1]], y=y)   # fit for treatment group
        model_ctrl = sklearn.module.model_class.fit(X = df[df[w==0]], y=y)  # fit for control group

        res = list(model_treatment = model_trn,
                model_control = model_ctrl, 
                model_class = model_class)

        return(res)

    def two_model_pred(self, models, X_pred, continuous = False):
        """
        Makes predicitons using the "Two Model Approach" (Double Model, Separate Model)

        Based on
        --------
        - "A Literature Survey and Experimental Evaluation of the State-of-the-Art in Uplift Modeling:
        A Stepping Stone Toward the Development of Prescriptive Analytics" (Devriendt, 2018) 
        - "Incremental Value Modeling" (Hansotia, 2002)

        Requirements
        ------------
        - NumPy : for arrays
        - scikit-learn : used for predictions

        Parameters
        ----------
        models : a list of two models that have been fit on the treatment and control groups respectively
        X_pred : new data on which to make a prediction

        Returns
        -------
        - A NumPy array of predicted outcomes for each unit in X_pred based on treatment assignment
        """
        import numpy as np

        pred_trt = models[model_treatment].predict(X_pred)
        pred_ctrl = models[model_control].predict(X_pred)

        tuples = [(pred_trt[i], pred_ctrl(i)) for i in list(range(len(X_pred)))]

        return np.array(tuples)